from __future__ import annotations

from typing import Any, Optional

from pydantic import BaseModel


class SettingsModel(BaseModel):
    version: float = 0.1
    ai_service: Optional[str] = None
    ai_model: Optional[str] = None
    ai_persona: Optional[str] = "You are a helpful assistant. Answer all questions to the best of your ability."
    ai_streaming: Optional[bool] = True
    ai_model_kwargs: Optional[dict[str, Any]] = {}
    ai_n: Optional[int] = 1
    ai_max_tokens: Optional[int] = 2048
    ai_history_depth: Optional[int] = 20



from __future__ import annotations

import json
import re
from typing import Any


def to_str(value: Any) -> str | None:
    return str(value) if value else None


def from_str(value: str or None) -> Any | None:
    return str(value) if str else ""


def str_to_num(value: float | str) -> float | str:
    if isinstance(value, (float, int)):
        return value
    if isinstance(value, str):
        try:
            result = int(value)
        except ValueError:
            try:
                result = float(value)
            except ValueError:
                result = value
        return result
    msg = f"Invalid type {type(value)}"
    raise TypeError(msg)


def str_to_dict(value: str) -> dict:
    val = re.sub(r"\s+", "", value)

    result = {}
    for kv in val.split(","):
        if "=" in kv:
            key, value_str = kv.split("=")
            if not key or not value_str:
                msg = "Invalid key-value pair"
                raise ValueError(msg)
            result[key] = str_to_num(value_str)

    return result


def dict_to_str(value: dict) -> str:
    if not isinstance(value, dict):
        msg = "Input must be a dictionary"
        raise TypeError(msg)

    sorted_items = sorted(value.items())

    def is_nested(d):
        return any(isinstance(v, dict) for v in d.values())

    filtered_items = [(key, val) for key, val in sorted_items if not is_nested({key: val})]

    return ",".join(f"{key}={json.dumps(val)}" for key, val in filtered_items)



import asyncio
from typing import TYPE_CHECKING, Dict, List

from paita.ai.bedrock import get_bedrock_client
from paita.ai.enums import AIService
from paita.utils.asyncify import asyncify
from paita.utils.logger import log

if TYPE_CHECKING:
    from openai.types import Model


async def list_models(ai_service: str = AIService.AWSBedRock.value):
    if ai_service == AIService.AWSBedRock.value:
        return await bedrock_models()
    if ai_service == AIService.OpenAIChatGPT.value:
        return await openai_models()
    # if ai_service == AIService.Mock:
    #     return [
    #         "mock_model_1",
    #         "mock_model_2"
    #     ]
    msg = f"Invalid value for {ai_service=}"
    raise ValueError(msg)


async def list_all_models() -> Dict[str, List[str]]:
    services = [service.value for service in AIService]
    tasks = [list_models(service) for service in services]
    responses: [[str]] = await asyncio.gather(*tasks)
    response: Dict[str, List[str]] = dict(zip(services, responses))
    # log.debug(f"{response=}")
    return response


@asyncify
def bedrock_models() -> [str]:
    try:
        bedrock_client = get_bedrock_client(runtime=False)
        response = bedrock_client.list_foundation_models(byOutputModality="TEXT")
        return [model["modelId"] for model in response["modelSummaries"]]
    except Exception as e:  # noqa: BLE001 TODO
        log.info(e)
        return []


@asyncify
def openai_models() -> [str]:
    from openai import OpenAI

    try:
        openai_client = OpenAI()
        response: [Model] = openai_client.models.list()
        return [model.id for model in response.data]
    except Exception as e:  # noqa: BLE001 TODO
        log.info(e)
        return []


async def main():
    log.debug(await bedrock_models())
    log.debug(await openai_models())
    log.debug(await list_all_models())


if __name__ == "__main__":
    asyncio.run(main())



import os
from typing import TYPE_CHECKING

from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.schema.output_parser import StrOutputParser
from langchain_community.chat_models import bedrock as br
from langchain_core.runnables.history import RunnableWithMessageHistory

from paita.ai.callbacks import AsyncHandler
from paita.ai.chat_history import ChatHistory
from paita.ai.mock_model import MockModel
from paita.ai.models import AIService
from paita.utils.logger import log
from paita.utils.settings_model import SettingsModel

if TYPE_CHECKING:
    from langchain_core.language_models import BaseChatModel


HISTORY_FILE_NAME = "chat_history"

# https://github.com/langchain-ai/langchain/issues/11668
BEDROCK_DISABLE_STREAMING = bool(os.getenv("BEDROCK_DISABLE_STREAMING", "True"))


class Chat:
    """
    Chat capsulates chat history and can use different AI Models
    """

    def __init__(self):
        self._model: BaseChatModel = None
        self._settings_model: SettingsModel = None
        self._callback_handler: AsyncHandler = None
        self.parser: StrOutputParser = StrOutputParser()

    def init_model(
        self,
        *,
        settings_model: SettingsModel,
        callback_handler: AsyncHandler,
    ):
        self._settings_model = settings_model
        self._callback_handler = callback_handler

        if settings_model.ai_service == AIService.AWSBedRock.value:
            model_kwargs = {
                # "max_tokens_to_sample": self._settings_model.ai_max_tokens,
                "temperature": 1,
                "top_k": 250,
                "top_p": 0.8,
                # "stop_sequences": ["\n\nHuman"],
            }
            if self._settings_model.ai_model_kwargs:
                model_kwargs.update(self._settings_model.ai_model_kwargs)
            log.debug(f"{model_kwargs=}")
            self._model = br.BedrockChat(
                model_id=settings_model.ai_model,
                streaming=(False if BEDROCK_DISABLE_STREAMING else self._settings_model.ai_streaming),
                model_kwargs=model_kwargs,
                # max_tokens=settings_model.ai_max_tokens,
                # n=settings_model.ai_n,
                callbacks=[callback_handler],
                # callback_manager=callback_handler,
            )
        elif settings_model.ai_service == AIService.OpenAIChatGPT.value:
            from langchain_openai import ChatOpenAI

            model_kwargs = {
                # "top_k": 250,
                "top_p": 0.8,
                "frequency_penalty": 0.8,
                "presence_penalty": 0.8,
                # "stop_sequences": ["\n\nHuman"],
            }
            temperature = 1
            if self._settings_model.ai_model_kwargs is not None:
                model_kwargs.update(self._settings_model.ai_model_kwargs)
                if "temperature" in model_kwargs:
                    temperature = int(model_kwargs["temperature"])
                    del model_kwargs["temperature"]
            log.debug(f"{model_kwargs=}")
            self._model = ChatOpenAI(
                model_name=settings_model.ai_model,
                streaming=settings_model.ai_streaming,
                model_kwargs=model_kwargs,
                max_tokens=settings_model.ai_max_tokens,
                temperature=temperature,
                # n=settings_model.ai_n,
                callbacks=[callback_handler],
                # callback_manager=callback_handler,
            )
        else:
            self._model = MockModel()

    async def request(self, data: str, *, chat_history: ChatHistory) -> str:
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    self._settings_model.ai_persona,
                ),
                MessagesPlaceholder(variable_name="chat_history"),
                ("human", "{input}"),
            ]
        )

        if self._settings_model.ai_streaming and not BEDROCK_DISABLE_STREAMING:
            chain = prompt | self._model | self.parser
            chain_with_message_history = RunnableWithMessageHistory(
                chain,
                lambda session_id: chat_history.history,  # noqa: ARG005
                input_messages_key="input",
                history_messages_key="chat_history",
            )
            async for _ in chain_with_message_history.astream(
                {"input": data},
                {"configurable": {"session_id": "unused"}},
            ):
                pass
        else:
            chain = prompt | self._model | self.parser
            chain_with_message_history = RunnableWithMessageHistory(
                chain,
                lambda session_id: chat_history.history,  # noqa: ARG005
                input_messages_key="input",
                history_messages_key="chat_history",
            )
            await chain_with_message_history.ainvoke(
                {"input": data},
                {"configurable": {"session_id": "unused"}},
            )
        await self._trim_history(chat_history, max_length=self._settings_model.ai_history_depth)

    @classmethod
    async def _trim_history(cls, chat_history: ChatHistory, *, max_length: int = 20):
        stored_messages = await chat_history.history.aget_messages()
        if len(stored_messages) <= max_length:
            return

        await chat_history.history.aclear()
        await chat_history.history.aadd_messages(stored_messages[-max_length:])

    async def _summarize_messages(self, chat_history: ChatHistory, *, max_length: int = 20):
        stored_messages = chat_history.history.messages
        if len(stored_messages) <= max_length:
            return
        summarization_prompt = ChatPromptTemplate.from_messages(
            [
                MessagesPlaceholder(variable_name="chat_history"),
                (
                    "user",
                    "Distill the above chat messages into a single summary message."
                    "Include as many specific details as you can.",
                ),
            ]
        )
        summarization_chain = summarization_prompt | self._model

        summary_message = summarization_chain.invoke({"chat_history": stored_messages})

        chat_history.history.clear()

        chat_history.history.add_message(summary_message)



import asyncio

from paita.utils.logger import log

markdown_content = [
    "# Sample Markdown Content\n",
    "\n",
    "This is a simple example of using Markdown in a list:\n",
    "\n",
    "- **Bold Text**: **This is bold text**\n",
    "- *Italic Text*: *This is italic text*\n",
    "- [Link](https://www.example.com): \n" "- Click [here](https://www.example.com) for more information\n",
    "- Bullet Points:\n",
    "  - Point 1\n",
    "  - Point 2\n",
    "- Numbered List:\n",
    "  1. First item\n",
    "  2. Second item\n",
    "\n",
    "Feel free to customize and add more elements to suit your needs!",
]


class MockModel:
    async def request(self, data: str) -> str:
        log.debug(data)
        if self.streaming:
            await asyncio.sleep(2)  # initial request sleep
            for item in markdown_content:
                self.callback_on_token(item)
                await asyncio.sleep(0.2)
        else:
            self.callback_on_token(markdown_content)
        self.callback_on_end()

    async def streaming(self):
        await asyncio.sleep(2)  # initial request sleep
        for item in markdown_content:
            self.callback_on_token(item)
            await asyncio.sleep(0.2)



import pytest

from paita.ai.models import list_all_models


# @pytest.mark.skip("Needs mocking or only integration test.")
@pytest.mark.asyncio
async def test_list_all_models():
    _ = await list_all_models()



from pathlib import PurePath
from typing import List, Tuple

from cache3 import DiskCache
from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, VerticalScroll
from textual.screen import ModalScreen
from textual.validation import Function, Number
from textual.widgets import Button, Checkbox, Header, Input, Select, TextArea

import paita.localization.labels as label
from paita.ai.enums import Tag
from paita.utils.logger import log
from paita.utils.settings_manager import SettingsManager, SettingsModel
from paita.utils.string_utils import dict_to_str, str_to_dict, str_to_num, to_str


class SettingsScreen(ModalScreen[bool]):
    CSS_PATH = PurePath(__file__).parent / "styles" / "settings_screen.tcss"

    def __init__(
        self,
        *,
        settings: SettingsManager,
        cache: DiskCache,
        allow_cancel: bool = False,
    ):
        super().__init__()
        self._settings: SettingsManager = settings
        self._model: SettingsModel = settings.model.model_copy()
        # log.debug(f"{self._model=}")
        self._cache: DiskCache = cache
        self._allow_cancel: bool = allow_cancel

        # Prefill select options based on stored settings values
        self._ai_services: List[Tuple[str, str]] = list(self._cache.keys(tag=Tag.AI_MODELS.value))
        if self._model.ai_service not in self._ai_services:
            self._model.ai_service = self._ai_services[0]

        self._ai_models: List[Tuple[str, str]] = self._cache.get(self._model.ai_service, [], tag=Tag.AI_MODELS.value)
        if self._model.ai_model not in self._ai_models:
            self._model.ai_model = self._ai_models[0]

    def compose(self) -> ComposeResult:
        yield Header()
        with Container(classes="settings"):
            with VerticalScroll(id="settings"):
                yield Select(
                    [(item, item) for item in self._ai_services],
                    value=self._model.ai_service,
                    prompt="AI Service",
                    id="ai_service",
                    classes="settings_option",
                    allow_blank=False,
                )
                yield Select(
                    [(item, item) for item in self._ai_models],
                    value=self._model.ai_model,
                    prompt="AI Model",
                    id="ai_model",
                    classes="settings_option",
                    allow_blank=False,
                )
                yield TextArea(
                    text=self._model.ai_persona,
                    id="ai_persona",
                    classes="settings_textarea",
                )
                yield Input(
                    placeholder=label.AI_MODEL_KWARGS,
                    value=dict_to_str(self._model.ai_model_kwargs),
                    id="ai_model_kwargs",
                    classes="settings_input",
                    validators=[Function(self.validate_ai_model_kwargs, "Invalid dictionary")],
                )
                yield Input(
                    placeholder=label.AI_MAX_TOKENS,
                    value=to_str(self._model.ai_max_tokens),
                    id="ai_max_tokens",
                    classes="settings_input",
                    type="integer",
                    validators=[Number(minimum=1, maximum=100000)],
                    # max_length=4,
                )
                yield Input(
                    placeholder=label.AI_HISTORY_DEPTH,
                    value=to_str(self._model.ai_history_depth),
                    id="ai_history_depth",
                    classes="settings_input",
                    type="integer",
                    validators=[Number(minimum=1, maximum=100)],
                    # max_length=2,
                )
                yield Input(
                    placeholder=label.AI_N,
                    value=to_str(self._model.ai_n),
                    id="ai_n",
                    classes="settings_input",
                    type="integer",
                    validators=[Number(minimum=1, maximum=10)],
                    # max_length=2,
                )
                yield Checkbox(
                    label.AI_STREAMING,
                    value=self._model.ai_streaming,
                    id="ai_streaming",
                    classes="settings_checkbox",
                )
                with Horizontal(classes="settings_block"):
                    yield Button(label="Apply", variant="success", id="apply")
                    if self._allow_cancel:
                        yield Button(label="Cancel", variant="warning", id="cancel")

    @staticmethod
    def validate_ai_model_kwargs(value: str):
        try:
            str_to_dict(value)
        except ValueError:
            return False
        else:
            return True

    async def on_mount(self) -> None:
        if self._model.ai_service is Select.BLANK or self._model.ai_model is Select.BLANK:
            self.query_one("#apply").disabled = True

    @on(Select.Changed)
    async def select_changed(self, event: Select.Changed) -> None:
        if event.control.id == "ai_service":
            models = self._cache.get(event.value, Select.BLANK, tag=Tag.AI_MODELS.value)
            widget: Select = self.query_one("#ai_model")
            widget.set_options((item, item) for item in models)
        elif event.control.id == "ai_model":
            self.query_one("#apply").disabled = False
        else:
            log.error(f"Undefined {event.id=}")

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "apply":
            model: SettingsModel = SettingsModel(
                ai_service=self.query_one("#ai_service").value,
                ai_model=self.query_one("#ai_model").value,
                # ai_persona=self.query_one("#ai_persona").text,
                # ai_streaming=self.query_one("#ai_streaming").value,
                # ai_model_kwargs=json.loads(self.query_one("#ai_model_kwargs").value),
                # ai_n=self.query_one("#ai_n").value,
                # ai_max_tokens=self.query_one("#ai_max_tokens").value,
                # ai_history_depth=self.query_one("#ai_history_depth").value,
            )
            if (value := self.query_one("#ai_persona").text) != "":
                model.ai_persona = value
            if (value := self.query_one("#ai_model_kwargs").value) != "":
                model.ai_model_kwargs = str_to_dict(value)
            if (value := self.query_one("#ai_n").value) != "":
                model.ai_n = str_to_num(value)
            if (value := self.query_one("#ai_max_tokens").value) != "":
                model.ai_max_tokens = str_to_num(value)
            if (value := self.query_one("#ai_history_depth").value) != "":
                model.ai_history_depth = str_to_num(value)
            model.ai_streaming = self.query_one("#ai_streaming").value

            # log.debug(f"{model=}")
            self._settings.model = model
            await self._settings.save()
            self.dismiss(True)
        else:
            self.dismiss(False)



import asyncio
from functools import partial, wraps


def asyncify(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)

    return run



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0
"""Helper utilities for working with Amazon Bedrock from Python notebooks"""
# Python Built-Ins:
from __future__ import annotations

import os
from typing import Optional

# External Dependencies:
import boto3
from botocore.config import Config

from paita.utils.logger import log


def get_bedrock_client(
    *,
    assumed_role: Optional[str] = None,
    region: Optional[str] = None,
    runtime: Optional[bool] = True,
):
    target_region = os.environ.get("AWS_REGION", os.environ.get("AWS_DEFAULT_REGION")) if region is None else region

    log.debug(f"Create new client\n  Using region: {target_region}")
    session_kwargs = {"region_name": target_region}
    client_kwargs = {**session_kwargs}

    profile_name = os.environ.get("AWS_PROFILE")
    if profile_name:
        log.debug(f"  Using profile: {profile_name}")
        session_kwargs["profile_name"] = profile_name

    retry_config = Config(
        region_name=target_region,
        connect_timeout=10,
        read_timeout=3,
        retries={
            "max_attempts": 3,
            "mode": "standard",
        },
    )
    session = boto3.Session(**session_kwargs)

    if assumed_role:
        log.debug(f"  Using role: {assumed_role}", end="")
        sts = session.client("sts")
        response = sts.assume_role(RoleArn=str(assumed_role), RoleSessionName="langchain-llm-1")
        log.debug(" ... successful!")
        client_kwargs["aws_access_key_id"] = response["Credentials"]["AccessKeyId"]
        client_kwargs["aws_secret_access_key"] = response["Credentials"]["SecretAccessKey"]
        client_kwargs["aws_session_token"] = response["Credentials"]["SessionToken"]
    service_name = "bedrock-runtime" if runtime else "bedrock"

    return session.client(service_name=service_name, config=retry_config, **client_kwargs)
